import math
import numpy as np
import os
from time import sleep, time
import glob
import random

from isaacgym import gymapi, gymutil, gymtorch
import torch

import utils

ASSET_ROOT = "assets/"

DEFAULT_POS = [0, 0.67, -1.25, 0, 0.67, -1.25, 0, 0.67, -1.25, 0, 0.67, -1.25]

# For global use
GYM = gymapi.acquire_gym()

class KinematicActor:
    def __init__(self, asset, env, actor_index):
        self.actor_index = actor_index

        self.asset = asset
        self.env = env

        # Flag to inidicate whether the object has currently loaded a reference
        # motion to reproduce.
        self.ref_motion_is_loaded = False
        # Stores the frames of the reference motion
        self.ref_motion_frames = None

        self.num_dofs = GYM.get_asset_dof_count(self.asset)

        self.handle = self._create_actor()

    def _create_actor(self):
        # Add actor
        pose = gymapi.Transform()
        pose.p = gymapi.Vec3(0.0, 0.0, 0.6)
        # 90 degree around X and -90 degree around Y. Makes the robot stand 
        # upright and face "forward" (positive X)
        pose.r = gymapi.Quat(0.5, -0.5, -0.5, 0.5)
        actor_handle = GYM.create_actor(self.env, self.asset, pose, 
                                        "laikago_kinematic", self.actor_index, 1)

        # Configure DOF properties
        props = GYM.get_actor_dof_properties(env, actor_handle)
        props["driveMode"][:] = gymapi.DOF_MODE_POS
        props["stiffness"][:] = 5000.0
        props["damping"][:] = 100.0

        GYM.set_actor_dof_properties(self.env, actor_handle, props)

        return actor_handle

    def load_reference_motion(self, dof_frames):
        """ Loads the reference motion from `dof_frames`. A reference motion
        consists of a sequence of frames that contains the root position (3)
        and orientation (4) of the character + the PD targets for each joint.

        """

        self.ref_motion_is_loaded = True
        self.ref_motion_frames = dof_frames
        
        self.motion_i = 0
        self.motion_length = len(self.ref_motion_frames)

    def load_next_frame(self):
        """ Reproduces the next frame of the reference motion. """

        # Load current frame
        frame = self.ref_motion_frames[self.motion_i]

        # This attribute is used externally by `set_body_pose_actors`.
        # `pose_target` for this actor is set using the tensor API (which sets
        # the value for ALL actors simultaneously).
        self.pose_target = frame[:7] # Position + orientation
        self.pd_targets = frame[7:]

        # Set PD targets to the robot not using the Tensor API. If 
        # `set_pd_targets_to_actors` which uses the Tensor API is called, the 
        # line below should be commented and vice-versa.
        #self.set_pd_targets(self.pd_targets)

        # Go to next frame
        self.motion_i += 1
        # Loop around the motion
        self.motion_i = self.motion_i % self.motion_length

    def set_pd_targets(self, pd_targets):
        """ Sets the DOF position targets to `pos_targets`. Makes sure that
        the length of the list and number of joints/dofs match.

        """

        assert len(pd_targets) == self.num_dofs

        success = GYM.set_actor_dof_position_targets(self.env, self.handle, pd_targets)
        print(f"Set PD targets: `{success}`")

def set_body_pose_to_actors(actors, root_tensor):
    # Reads the pose from current frame from each actor
    poses = [a.pose_target for a in actors]
    # Note: the tensor below doesn't need `.cuda()` when running on GPU since 
    # its values are copied to `root_tensor` (which already is in GPU).
    poses = torch.tensor(poses, dtype=torch.float32) 

    # Sets the position and orientation entries of the root tensor
    root_tensor[:, :7] = poses

    # Unwraps tensor and set it to all actors
    GYM.set_actor_root_state_tensor(SIM, gymtorch.unwrap_tensor(root_tensor))

def set_body_pose_to_actors_fixed_at_origin(actors, root_tensor):
    # Sets the pose to be fixed at origin for every actor
    poses = [[0.0, 0.0, 0.6, 0.5, -0.5, -0.5, 0.5] for a in actors]
    poses = torch.tensor(poses, dtype=torch.float32)

    # Sets the position and orientation entries of the root tensor
    root_tensor[:, :7] = poses

    # Unwraps tensor and set it to all actors
    GYM.set_actor_root_state_tensor(SIM, gymtorch.unwrap_tensor(root_tensor))

def set_dof_state_to_actors(actors, dof_state_tensor):
    pd_targets = [a.pd_targets for a in actors]
    pd_targets = torch.tensor(pd_targets, dtype=torch.float32).cuda()

    dof_state_tensor[..., 0] = pd_targets

    GYM.set_dof_state_tensor(SIM, gymtorch.unwrap_tensor(dof_state_tensor))

def set_pd_targets_to_actors(actors):
    pd_targets = [a.pd_targets for a in actors]
    pd_targets = torch.tensor(pd_targets, dtype=torch.float32).cuda()

    GYM.set_dof_position_target_tensor(SIM, gymtorch.unwrap_tensor(pd_targets))

# Set up simulation parameters
sim_params = gymapi.SimParams()
sim_params.dt = dt = 1.0 / 60.0
sim_params.up_axis = gymapi.UP_AXIS_Z
sim_params.gravity = gymapi.Vec3(0.0, 0.0, -9.8)

# Using GPU or not...
sim_params.use_gpu_pipeline = True
sim_params.physx.use_gpu = True

# Create simulation
SIM = GYM.create_sim(0, 0, gymapi.SIM_PHYSX, sim_params)

# Load Laikago asset
asset_fn = "laikago_description/laikago_toes_limits.urdf"
asset_options = gymapi.AssetOptions()
asset_options.fix_base_link = True
asset = GYM.load_asset(SIM, ASSET_ROOT, asset_fn, asset_options)

# Define ground plane parameters
plane_params = gymapi.PlaneParams()
plane_params.normal = gymapi.Vec3(0, 0, 1) # z-up!
GYM.add_ground(SIM, plane_params)

# Create viewer window
viewer = GYM.create_viewer(SIM, gymapi.CameraProperties())

# Load the reference motions from .txt files
motion_txts = glob.glob("motions/*.txt")
print(f"Loaded reference motions: `{motion_txts}`")
ref_motions = [utils.load_reference_motion(fn) for fn in motion_txts]

# Create envs and actors
num_envs = 1
num_per_row = 8
spacing = 2.0
env_lower = gymapi.Vec3(-spacing, 0.0, -spacing)
env_upper = gymapi.Vec3(spacing, spacing, spacing)

envs = []
actors = []

for i in range(num_envs):
    env = GYM.create_env(SIM, env_lower, env_upper, num_per_row)
    envs.append(env)

    # Create actor.
    kin_actor = KinematicActor(asset, env, actor_index=i)

    # Load motion into actor
    rand_i = random.randint(0, len(ref_motions)-1)
    rand_i = 2 # Fixing a motion
    motion_frames = np.array(ref_motions[rand_i]["Frames"], dtype=np.float32)
    kin_actor.load_reference_motion(motion_frames)

    actors.append(kin_actor)

##### Run Simulation #####
GYM.prepare_sim(SIM)

# Root tensor for all actors. Used to manipulate root position and orientation
_root_tensor = GYM.acquire_actor_root_state_tensor(SIM)
root_tensor = gymtorch.wrap_tensor(_root_tensor)

# Look at the first env
cam_pos = gymapi.Vec3(0.7, 1.5, 0.7)
cam_target = gymapi.Vec3(0.5, 0.0, 0)
GYM.viewer_camera_look_at(viewer, None, cam_pos, cam_target)

ticks = 0
while not GYM.query_viewer_has_closed(viewer):
    GYM.simulate(SIM)
    GYM.fetch_results(SIM, True)

    GYM.refresh_actor_root_state_tensor(SIM)
    GYM.refresh_dof_state_tensor(SIM)

    if ticks > 0:
        start = time()

        # Load the next frame for each actor
        for a in actors:
            a.load_next_frame()

        set_pd_targets_to_actors(actors)

        set_body_pose_to_actors(actors, root_tensor)

        #set_body_pose_to_actors_fixed_at_origin(actors, root_tensor)

        #set_dof_state_to_actors(actors, dof_state_tensor)

        elapsed = time() - start

        print(f"Action loop took {elapsed*1000:.2f} ms")

    GYM.step_graphics(SIM)
    GYM.draw_viewer(viewer, SIM, True)

    # Render in real time
    GYM.sync_frame_time(SIM)

    ticks += 1
